import requests
import time
import logging

logger = logging.getLogger(__name__)


class PandoraStatsReader(object):
    # TODO: maybe make stats collection asyncronous
    def next(self):
        try:
            pandora_response = requests.get("http://localhost:1234/debug/vars")
            pandora_stat = pandora_response.json()

            return [{'ts': int(time.time() - 1),
                     'metrics': {
                         'instances':
                         pandora_stat.get("engine_ActiveRequests"),
                         'reqps': pandora_stat.get("engine_ReqPS"),
                     }}]
        except requests.ConnectionError:
            logger.info("Pandora expvar http interface is unavailable")
        except requests.HTTPError:
            logger.warning("Pandora expvar http interface is unavailable",
                           exc_info=True)
        except Exception:
            logger.warning("Couldn't decode pandora stat:\n%s\n",
                           pandora_response.text,
                           exc_info=True)

        return [{'ts': int(time.time() - 1),
                 'metrics': {'instances': 0,
                             'reqps': 0}}]

    def close(self):
        pass

    def __iter__(self):
        return self



# TODO: make the next two lines unnecessary
# pylint: disable=line-too-long
# pylint: disable=missing-docstring
import json
import logging
import os

from ..Aggregator import Plugin as AggregatorPlugin
from ..Monitoring import Plugin as MonitoringPlugin
from ..Telegraf import Plugin as TelegrafPlugin
from ...common.interfaces import AbstractPlugin, MonitoringDataListener, AggregateResultListener

logger = logging.getLogger(__name__)  # pylint: disable=C0103


class Plugin(AbstractPlugin, AggregateResultListener, MonitoringDataListener):
    # pylint:disable=R0902
    SECTION = 'json_report'

    def __init__(self, core):
        super(Plugin, self).__init__(core)
        self._is_telegraf = None

    def get_available_options(self):
        return ['monitoring_log', 'test_data_log', 'test_stats_log']

    def configure(self):
        self.monitoring_logger = self.create_file_logger('monitoring',
                                                         self.get_option('monitoring_log',
                                                                         'monitoring.log'))
        self.aggregator_data_logger = self.create_file_logger('aggregator_data',
                                                              self.get_option('test_data_log',
                                                                              'test_data.log'))
        self.core.job.subscribe_plugin(self)

    def create_file_logger(self, logger_name, file_name, formatter=None):
        loggr = logging.getLogger(logger_name)
        loggr.setLevel(logging.INFO)
        handler = logging.FileHandler(os.path.join(self.core.artifacts_dir, file_name), mode='w')
        handler.setLevel(logging.INFO)
        if formatter:
            handler.setFormatter(formatter)
        loggr.addHandler(handler)
        loggr.propagate = False
        return loggr

    def on_aggregated_data(self, data, stats):
        """
        @data: aggregated data
        @stats: stats about gun
        """
        self.aggregator_data_logger.info(json.dumps({'data': data, 'stats': stats}))

    def monitoring_data(self, data_list):
        if self.is_telegraf:
            self.monitoring_logger.info(json.dumps(data_list))
        else:
            [self.monitoring_logger.info(data.strip()) for data in data_list if data]

    @property
    def is_telegraf(self):
        if self._is_telegraf is None:
            self._is_telegraf = 'Telegraf' in self.core.job.monitoring_plugin.__module__
        return self._is_telegraf


#!/usr/bin/env python
""" The agent bundle, contains all metric classes and agent running code """
import subprocess
import logging
import os
import sys
import signal
import ConfigParser
import json
import threading
import time

from optparse import OptionParser
import Queue as q



logger = logging.getLogger("agent")
collector_logger = logging.getLogger("telegraf")


def signal_handler(sig, frame):
    """ required for non-tty python runs to interrupt """
    raise KeyboardInterrupt()

signal.signal(signal.SIGINT, signal_handler)
signal.signal(signal.SIGTERM, signal_handler)


class DataReader(object):
    """generator reads from source line-by-line"""
    def __init__(self, filename, pipe=False):
        self.pipe = pipe
        if not self.pipe:
            self.monout = open(filename, 'r')
        else:
            self.monout = filename
        self.buffer = ""
        self.closed = False

    def __iter__(self):
        while not self.closed:
            data = self.monout.readline()
            if data:
                parts = data.rsplit('\n', 1)
                if len(parts) > 1:
                    ready_chunk = self.buffer + parts[0] + '\n'
                    self.buffer = parts[1]
                    yield ready_chunk
                else:
                    self.buffer += parts[0]
            else:
                time.sleep(1)
        if not self.pipe:
            self.monout.close()

    def close(self):
        self.closed = True


class Consolidator(object):
    """generator consolidates data from source, cache it by timestamp"""
    def __init__(self, source):
        self.source = source
        self.results = {}

    def __iter__(self):
        for chunk in self.source:
            #logger.info('Consolidator got %s', chunk)
            if chunk:
                try:
                    data = json.loads(chunk)
                except ValueError:
                    logger.error('unable to decode chunk %s', chunk, exc_info=True)
                else:
                    try:
                        ts = data['timestamp']
                        self.results.setdefault(ts, {})
                        for key, value in data['fields'].iteritems():
                            if data['name'] == 'diskio':
                                data['name'] = "{metric_name}-{disk_id}".format(
                                    metric_name=data['name'],
                                    disk_id=data['tags']['name']
                                )
                            elif data['name'] == 'net':
                                data['name'] = "{metric_name}-{interface}".format(
                                    metric_name=data['name'],
                                    interface=data['tags']['interface']
                                )
                            elif data['name'] == 'cpu':
                                data['name'] = "{metric_name}-{cpu_id}".format(
                                    metric_name=data['name'],
                                    cpu_id=data['tags']['cpu']
                                )
                            key = data['name']+"_"+key
                            if key.endswith('_exec_value'):
                                key = key.replace('_exec_value','')
                            self.results[ts][key] = value
                    except KeyError:
                        logger.error('Malformed json from source: %s', chunk, exc_info=True)
                    except:
                        logger.error('Something nasty happend in consolidator work', exc_info=True)
            if len(self.results) > 5:
                ready_to_go_index = min(self.results)
                yield json.dumps(
                    {
                        ready_to_go_index: self.results.pop(ready_to_go_index, None)
                    }
                )



class Drain(threading.Thread):
    """
    Drain a generator to a destination that answers to put(), in a thread
    """

    def __init__(self, source, destination):
        super(Drain, self).__init__()
        self.source = source
        self.destination = destination
        self._finished = threading.Event()
        self._interrupted = threading.Event()

    def run(self):
        for item in self.source:
            self.destination.put(item)
            if self._interrupted.is_set():
                break
        self._finished.set()

    def wait(self, timeout=None):
        self._finished.wait(timeout=timeout)

    def close(self):
        self._interrupted.set()


class AgentWorker(threading.Thread):
    def __init__(self, telegraf_path):
        super(AgentWorker, self).__init__()
        self.working_dir = os.path.dirname(__file__)
        self.startups = []
        self.startup_processes = []
        self.shutdowns = []
        self.daemon = True  # Thread auto-shutdown
        self.finished = False
        self.drain = None
        self.drain_stdout = None
        self.drain_err = None
        self.data_reader = None
        self.telegraf_path = telegraf_path
        self.results = q.Queue()
        self.results_stdout = q.Queue()
        self.results_err = q.Queue()

    @staticmethod
    def popen(cmnd):
        return subprocess.Popen(
            cmnd,
            bufsize=0,
            preexec_fn=os.setsid,
            close_fds=True,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
        )

    def read_startup_config(self, cfg_file='agent_startup.cfg'):
        try:
            config = ConfigParser.ConfigParser()
            with open(os.path.join(self.working_dir, cfg_file), 'rb') as f:
                config.readfp(f)

            if config.has_section('startup'):
                for option in config.options('startup'):
                    if option.startswith('cmd'):
                        self.startups.append(config.get('startup', option))

            if config.has_section('shutdown'):
                for option in config.options('shutdown'):
                    if option.startswith('cmd'):
                        self.shutdowns.append(config.get('shutdown', option))
            logger.info(
                'Successfully loaded startup config.\n'
                'Startups: %s\n'
                'Shutdowns: %s\n', self.startups, self.shutdowns
            )
        except:
            logger.error(
                'Error trying to read agent startup config',
                exc_info=True)

    def run(self):
        logger.info("Running startup commands")
        for cmnd in self.startups:
            logger.debug("Run: %s", cmnd)
            proc = self.popen(cmnd)
            self.startup_processes.append(proc)

        logger.info('Starting metrics collector..')
        cmnd = "{telegraf} -config {working_dir}/agent.cfg".format(
            telegraf=self.telegraf_path,
            working_dir=self.working_dir
        )
        self.collector = self.popen(cmnd)


        telegraf_output = self.working_dir+'/monitoring.rawdata'

        for _ in range(10):
            logger.info("Waiting for telegraf...")
            if os.path.isfile(telegraf_output):
                break
            time.sleep(1)

        self.drain = Drain(
            Consolidator(
                DataReader(telegraf_output)
            ),
            self.results
        )
        self.drain.start()

        self.drain_stdout = Drain(
            DataReader(self.collector.stdout, pipe=True),
            self.results_stdout
        )
        self.drain_stdout.start()

        self.drain_err = Drain(
            DataReader(self.collector.stderr, pipe=True),
            self.results_err
        )
        self.drain_err.start()

        while not self.finished:
            for _ in range(self.results.qsize()):
                try:
                    data = self.results.get_nowait()
                    logger.debug('send %s bytes of data to collector', len(data))
                    sys.stdout.write(
                        str(data)+'\n'
                    )
                except q.Empty:
                    break
                except:
                    logger.error('Something nasty happend trying to send data', exc_info=True)
            for _ in range(self.results_stdout.qsize()):
                try:
                    data = self.results_stdout.get_nowait()
                    if data:
                        collector_logger.info("STDOUT: %s", data)
                except q.Empty:
                    break
            for _ in range(self.results_err.qsize()):
                try:
                    data = self.results_err.get_nowait()
                    if data:
                        collector_logger.info("STDERR: %s", data)
                except q.Empty:
                    break
            time.sleep(1)

        self.stop()

    def stop(self):
        logger.info("Terminating startup commands")
        for proc in self.startup_processes:
            logger.debug("Terminate: %s", proc)
            os.killpg(proc.pid, signal.SIGTERM)

        logger.info("Running shutdown commands")
        for cmnd in self.shutdowns:
            logger.debug("Run: %s", cmnd)
            subprocess.call(cmnd, shell=True)

        logger.info('Terminating collector process: %s', self.collector)
        if not self.collector.returncode:
            self.collector.terminate()
            self.collector.wait()
        self.finished = True
        sys.stderr.write('stopped\n')
        logger.info('Stopped via stdin')

        logger.info('retcode: %s', self.collector.returncode)

        logger.info("Worker thread finished")


def main():
    fname = os.path.dirname(__file__) + "/_agent.log"
    logging.basicConfig(
        level=logging.DEBUG, filename=fname,
        format='%(asctime)s [%(levelname)s] %(name)s:%(lineno)d %(message)s')

    parser = OptionParser()
    parser.add_option(
        "", "--telegraf", dest="telegraf_path",
        help="telegraf_path",
        default="/tmp/telegraf")
    parser.add_option(
        "", "--host", dest="hostname_path",
        help="telegraf_path",
        default="/usr/bin/telegraf")
    (options, args) = parser.parse_args()

    logger.info('Init')
    customs_script = os.path.dirname(__file__)+'/agent_customs.sh'
    try:
        logger.info('Trying to make telegraf executable: %s', options.telegraf_path)
        os.chmod(options.telegraf_path, 493)  # 0o755 compatible with old python versions. 744 is NOT enough
    except OSError:
        logger.warning(
            'Unable to set %s access rights to execute.',
            options.telegraf_path, exc_info=True)
    try:
        logger.info('Trying to make customs script executable: %s', customs_script)
        os.chmod(customs_script, 493)  # 0o755 compatible with old python versions. 744 is NOT enough
    except OSError:
        logger.warning(
            'Unable to set %s access rights to execute.',
            customs_script, exc_info=True)


    worker = AgentWorker(options.telegraf_path)
    worker.read_startup_config()

    logger.info('Starting AgentWorker: %s', worker)
    worker.start()

    try:
        logger.debug("Check for any stdin command for shutdown")
        cmd = sys.stdin.readline()
        if cmd:
            logger.info("Stdin cmd received: %s", cmd)
            worker.stop()
    except KeyboardInterrupt:
        logger.debug("Interrupted")
        worker.stop()


if __name__ == '__main__':
    main()



"""Module provides target monitoring
metrics collector - influxdata's `telegraf` - https://github.com/influxdata/telegraf/
backward compatibility with yandextank's Monitoring module configuration and tools.
"""

import datetime
import fnmatch
import json
import logging
import os
import time

import sys
if sys.version_info[0] < 3:
    from ConfigParser import NoOptionError
else:
    from configparser import NoOptionError

from ...common.resource import manager as resource
from ...common.interfaces import MonitoringDataListener, AbstractPlugin, AbstractInfoWidget
from ...common.util import expand_to_seconds

from ..Autostop import Plugin as AutostopPlugin, AbstractCriterion
from ..Console import Plugin as ConsolePlugin
from ..Telegraf.collector import MonitoringCollector

logger = logging.getLogger(__name__)


class Plugin(AbstractPlugin):
    """  resource mon plugin  """

    SECTION = 'telegraf' # may be redefined to 'monitoring' sometimes.

    def __init__(self, core):
        super(Plugin, self).__init__(core)
        self.jobno = None
        self.default_target = None
        self.default_config = "{path}/config/monitoring_default_config.xml".format(
            path=os.path.dirname(__file__)
        )
        self.config = None
        self.process = None
        self.monitoring = MonitoringCollector()
        self.die_on_fail = True
        self.data_file = None
        self.mon_saver = None


    @staticmethod
    def get_key():
        return __file__

    def start_test(self):
        if self.monitoring:
            self.monitoring.load_start_time = time.time()
            logger.debug("load_start_time = %s", self.monitoring.load_start_time)

    def get_available_options(self):
        return ["config", "default_target", "ssh_timeout"]

    def __detect_configuration(self):
        """
        we need to be flexible in order to determine which plugin's configuration
        specified and make appropriate configs to metrics collector

        :return: SECTION name or None for defaults
        """
        try:
            is_telegraf = self.core.get_option('telegraf', "config", None)
        except NoOptionError:
            is_telegraf = None
        try:
            is_monitoring = self.core.get_option('monitoring', "config", None)
        except NoOptionError:
            is_monitoring = None

        if is_telegraf and is_monitoring:
            raise ValueError('Both telegraf and monitoring configs specified. '
                               'Clean up your config and delete one of them')
        if is_telegraf and not is_monitoring:
            return 'telegraf'
        if not is_telegraf and is_monitoring:
            return 'monitoring'
        if not is_telegraf and not is_monitoring:
            # defaults target logic
            try:
                is_telegraf_dt = self.core.get_option('telegraf', "default_target", None)
            except NoOptionError:
                is_telegraf_dt = None
            try:
                is_monitoring_dt = self.core.get_option('monitoring', "default_target", None)
            except:
                is_monitoring_dt = None
            if is_telegraf_dt and is_monitoring_dt:
                raise ValueError('Both telegraf and monitoring default targets specified. '
                                   'Clean up your config and delete one of them')
            if is_telegraf_dt and not is_monitoring_dt:
                return
            if not is_telegraf_dt and is_monitoring_dt:
                self.core.set_option("telegraf", "default_target", is_monitoring_dt)
            if not is_telegraf_dt and not is_monitoring_dt:
                return

    def configure(self):
        self.detected_conf = self.__detect_configuration()
        if self.detected_conf:
            logging.info('Detected monitoring configuration: %s', self.detected_conf)
            self.SECTION = self.detected_conf
        self.config = self.get_option("config", "auto").strip()
        self.default_target = self.get_option("default_target", "localhost")
        if self.config.lower() == "none":
            self.monitoring = None
            self.die_on_fail = False
            return

        # FIXME [legacy] backward compatibility with Monitoring module configuration below.
        self.monitoring.ssh_timeout = expand_to_seconds(self.get_option("ssh_timeout", "5s"))

        # FIXME [legacy] handle raw XML config in .ini file
        if self.config[0] == "<":
            config_contents = self.config
        # handle http/https url or file path
        else:
            if self.config.lower() == "auto":
                self.die_on_fail = False
                with open(resource.resource_filename(self.default_config), 'rb') as def_config:
                    config_contents = def_config.read()
            else:
                with open(resource.resource_filename(self.config), 'rb') as config:
                    config_contents = config.read()

        # dump config contents into a file
        xmlfile = self.core.mkstemp(".xml", "monitoring_")
        self.core.add_artifact_file(xmlfile)
        with open(xmlfile, "wb") as f: # output file should be in binary mode to support py3
            f.write(config_contents)
        self.config = xmlfile

        try:
            autostop = self.core.get_plugin_of_type(AutostopPlugin)
            autostop.add_criterion_class(MetricHigherCriterion)
            autostop.add_criterion_class(MetricLowerCriterion)
        except KeyError:
            logger.debug(
                "No autostop plugin found, not adding instances criterion")

    def prepare_test(self):
        if not self.config or self.config.lower() == 'none':
            return

        if "Phantom" in self.core.job.generator_plugin.__module__:
            phantom = self.core.job.generator_plugin
            if phantom.phout_import_mode:
                logger.info("Phout import mode, disabling monitoring")
                self.config = None
                self.monitoring = None

            info = phantom.get_info()
            if info:
                self.default_target = info.address
                logger.debug("Changed monitoring target to %s", self.default_target)

        logger.info("Starting monitoring with config: %s", self.config)
        self.core.add_artifact_file(self.config, True)
        self.monitoring.config = self.config
        if self.default_target:
            self.monitoring.default_target = self.default_target

        # FIXME json report already save this artifact, fix pls
        self.data_file = self.core.mkstemp(".data", "monitoring_overall_")
        self.mon_saver = SaveMonToFile(self.data_file)
        self.monitoring.add_listener(self.mon_saver)
        self.core.add_artifact_file(self.data_file)

        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except Exception as ex:
            logger.debug("Console not found: %s", ex)
            console = None
        if console:
            widget = MonitoringWidget(self)
            console.add_info_widget(widget)
            self.monitoring.add_listener(widget)

        try:
            self.monitoring.prepare()
            self.monitoring.start()
            count = 0
            while not self.monitoring.first_data_received and count < 15 * 5:
                time.sleep(0.2)
                self.monitoring.poll()
                count += 1
        except:
            logger.error("Could not start monitoring", exc_info=True)
            if self.die_on_fail:
                raise
            else:
                self.monitoring = None

    def is_test_finished(self):
        if self.monitoring:
            data_len = self.monitoring.poll()
            logger.debug("Monitoring got %s lines", data_len)
        return -1

    def end_test(self, retcode):
        logger.info("Finishing monitoring")
        if self.monitoring:
            self.monitoring.stop()
            for log in self.monitoring.artifact_files:
                self.core.add_artifact_file(log)

            while self.monitoring.send_data:
                logger.info("Sending monitoring data rests...")
                self.monitoring.send_collected_data()
        if self.mon_saver:
            self.mon_saver.close()
        return retcode


class SaveMonToFile(MonitoringDataListener):
    """
    Default listener - saves data to file
    """

    def __init__(self, out_file):
        MonitoringDataListener.__init__(self)
        if out_file:
            self.store = open(out_file, 'w')

    def monitoring_data(self, data):
        self.store.write(json.dumps(data))
        self.store.write('\n')
        self.store.flush()

    def close(self):
        """ close open files """
        logger.debug("Closing monitoring file")
        if self.store:
            self.store.close()


class MonitoringWidget(AbstractInfoWidget, MonitoringDataListener):
    """
    Screen widget
    """

    def __init__(self, owner):
        AbstractInfoWidget.__init__(self)
        self.owner = owner
        self.data = {}
        self.sign = {}
        self.time = {}
        self.max_metric_len = 0

    def get_index(self):
        return 50

    def __handle_data_items(self, host, data):
        """ store metric in data tree and calc offset signs

        sign < 0 is CYAN, means metric value is lower then previous,
        sign > 1 is YELLOW, means metric value is higher then prevoius,
        sign == 0 is WHITE, means initial or equal metric value
        """
        for metric, value in data.iteritems():
            if value == '':
                self.sign[host][metric] = -1
                self.data[host][metric] = value
            else:
                if not self.data[host].get(metric, None):
                    self.sign[host][metric] = 1
                elif float(value) > float(self.data[host][metric]):
                    self.sign[host][metric] = 1
                elif float(value) < float(self.data[host][metric]):
                    self.sign[host][metric] = -1
                else:
                    self.sign[host][metric] = 0
                self.data[host][metric] = "%.2f" % float(value)


    def monitoring_data(self, block):
        # block sample :
        # [{'timestamp': 1480536634,
        #   'data': {
        #     'some.hostname.tld': {
        #       'comment': '',
        #       'metrics': {
        #         'custom:diskio_reads': 0, 'Net_send': 9922, 'CPU_steal': 0, 'Net_recv': 8489
        #       }
        #     }
        #   },
        #   ...
        # }]
        for chunk in block:
            host = chunk['data'].keys()[0]
            self.time[host] = chunk['timestamp']
            # if initial call, we create dicts w/ data and `signs`
            # `signs` used later to paint metrics w/ different colors
            if not self.data.get(host, None):
                self.data[host] = {}
                self.sign[host] = {}
                for key, value in chunk['data'][host]['metrics'].iteritems():
                    self.sign[host][key] = 0
                    self.data[host][key] = value
            else:
                self.__handle_data_items(host, chunk['data'][host]['metrics'])


    def render(self, screen):
        if not self.owner.monitoring:
            return "Monitoring is " + screen.markup.RED + "offline" + screen.markup.RESET
        else:
            res = "Monitoring is " + screen.markup.GREEN + \
                "online" + screen.markup.RESET + ":\n"
            for hostname, metrics in self.data.items():
                tm_stamp = datetime.datetime.fromtimestamp(float(self.time[
                    hostname])).strftime('%H:%M:%S')
                res += ("   " + screen.markup.CYAN + "%s" + screen.markup.RESET
                        + " at %s:\n") % (hostname, tm_stamp)
                for metric, value in sorted(metrics.iteritems()):
                    if self.sign[hostname][metric] > 0:
                        value = screen.markup.YELLOW + value + screen.markup.RESET
                    elif self.sign[hostname][metric] < 0:
                        value = screen.markup.CYAN + value + screen.markup.RESET
                    res += "      %s%s: %s\n" % (
                        ' ' * (self.max_metric_len - len(metric)),
                        metric.replace('custom:', '').replace('_', ' '), value)

            return res.strip()


class AbstractMetricCriterion(AbstractCriterion, MonitoringDataListener):
    """ Parent class for metric criterion """

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        try:
            self.mon = autostop.core.get_plugin_of_type(Plugin)
            if self.mon.monitoring:
                self.mon.monitoring.add_listener(self)
        except KeyError:
            logger.warning("No monitoring module, mon autostop disabled")
        self.triggered = False
        self.autostop = autostop

        self.host = param_str.split(',')[0].strip()
        self.metric = param_str.split(',')[1].strip()
        self.value_limit = float(param_str.split(',')[2])
        self.seconds_limit = expand_to_seconds(param_str.split(',')[3])
        self.last_second = None
        self.seconds_count = 0

    def monitoring_data(self, block):
        if self.triggered:
            return

        for chunk in block:
            host = chunk['data'].keys()[0]
            data = chunk['data'][host]['metrics']

            if not fnmatch.fnmatch(host, self.host):
                continue

            # some magic, converting custom metric names into names that was in config
            for metric_name in data.keys():
                if metric_name.startswith('custom:'):
                    config_metric_name = metric_name.replace('custom:', '')
                    data[config_metric_name] = data.pop(metric_name)

            if self.metric not in data.keys() or not data[self.metric]:
                data[self.metric] = 0
            logger.debug("Compare %s %s/%s=%s to %s", self.get_type_string(),
                         host, self.metric, data[self.metric],
                         self.value_limit)
            if self.comparison_fn(float(data[self.metric]), self.value_limit):
                if not self.seconds_count:
                    self.cause_second = self.last_second

                logger.debug(self.explain())

                self.seconds_count += 1
                if self.seconds_count >= self.seconds_limit:
                    logger.debug("Triggering autostop")
                    self.triggered = True
                    return
            else:
                self.seconds_count = 0

    def notify(self, data, stat):
        if self.seconds_count:
            self.autostop.add_counting(self)

        self.last_second = (data, stat)
        return self.triggered

    def comparison_fn(self, arg1, arg2):
        """ comparison function """
        raise NotImplementedError()


class MetricHigherCriterion(AbstractMetricCriterion):
    """ trigger if metric is higher than limit """

    def __init__(self, autostop, param_str):
        AbstractMetricCriterion.__init__(self, autostop, param_str)

    def get_rc(self):
        return 31

    @staticmethod
    def get_type_string():
        return 'metric_higher'

    def explain(self):
        items = (self.host, self.metric, self.value_limit, self.seconds_count)
        return "%s/%s metric value is higher than %s for %s seconds" % items

    def widget_explain(self):
        items = (self.host, self.metric, self.value_limit, self.seconds_count,
                 self.seconds_limit)
        return "%s/%s > %s for %s/%ss" % items, float(
            self.seconds_count) / self.seconds_limit

    def comparison_fn(self, arg1, arg2):
        return arg1 > arg2


class MetricLowerCriterion(AbstractMetricCriterion):
    """ trigger if metric is lower than limit """

    def __init__(self, autostop, param_str):
        AbstractMetricCriterion.__init__(self, autostop, param_str)

    def get_rc(self):
        return 32

    @staticmethod
    def get_type_string():
        return 'metric_lower'

    def explain(self):
        items = (self.host, self.metric, self.value_limit, self.seconds_count)
        return "%s/%s metric value is lower than %s for %s seconds" % items

    def widget_explain(self):
        items = (self.host, self.metric, self.value_limit, self.seconds_count,
                 self.seconds_limit)
        return "%s/%s < %s for %s/%ss" % items, float(
            self.seconds_count) / self.seconds_limit

    def comparison_fn(self, arg1, arg2):
        return arg1 < arg2



"""Target monitoring via SSH"""
import base64
import getpass
import logging
import os.path
import re
import sys
if sys.version_info[0] < 3:
    import ConfigParser
else:
    import configparser as ConfigParser
import tempfile
import time
from collections import defaultdict
from xml.etree import ElementTree as etree

from ...common.util import SecuredShell
from ...common.interfaces import MonitoringDataListener

logger = logging.getLogger(__name__)
logging.getLogger("paramiko.transport").setLevel(logging.WARNING)


def parse_xml(config):
    if os.path.exists(config):
        return etree.parse(config)
    else:
        return etree.fromstring(config)


class Config(object):
    """Config reader helper"""

    def __init__(self, config):
        self.tree = parse_xml(config)

    def loglevel(self):
        """Get log level from config file. Possible values: info, debug"""
        log_level = 'info'
        log_level_raw = self.tree.getroot().get('loglevel')
        if log_level_raw in ('info', 'debug'):
            log_level = log_level_raw
        return log_level


class AgentClient(object):
    """Agent client connection"""

    def __init__(self, adr, timeout):
        self.run = []
        self.host = adr['host']
        self.username = adr['username']
        self.python = adr['python']
        self.metric = adr['metric']
        self.port = adr['port']
        self.interval = adr['interval']
        self.custom = adr['custom']
        self.startups = adr['startups']
        self.shutdowns = adr['shutdowns']
        self.session = None
        self.buffer = ""
        self.ssh = SecuredShell(self.host, self.port, self.username, timeout)

        handle, cfg_path = tempfile.mkstemp('.cfg', 'agent_')
        os.close(handle)
        self.path = {
            # Destination path on remote host
            'AGENT_REMOTE_FOLDER': '/var/tmp/lunapark_monitoring',

            # Source path on tank
            'AGENT_LOCAL_FOLDER': os.path.dirname(__file__) + '/agent',
            'METRIC_LOCAL_FOLDER': os.path.dirname(__file__) + '/agent/metric',

            # Temp config path
            'TEMP_CONFIG': cfg_path
        }

    def start(self):
        """Start remote agent"""
        logger.debug('Start monitoring: %s', self.host)
        self.session = self.ssh.async_session(" ".join([
            "DEBUG=1", self.python, self.path['AGENT_REMOTE_FOLDER'] +
            '/agent.py', '-c', self.path['AGENT_REMOTE_FOLDER'] + '/agent.cfg',
            '-t', str(int(time.time()))
        ]))
        return self.session

    def read_maybe(self):
        chunk = self.session.read_maybe()
        if chunk:
            parts = chunk.rsplit('\n', 1)
            if len(parts) > 1:
                ready_chunk = self.buffer + parts[0] + '\n'
                self.buffer = parts[1]
                return ready_chunk
            else:
                self.buffer += parts[0]
                return None
        return None

    def create_agent_config(self, loglevel):
        """Creating config"""
        try:
            float(self.interval)
        except:
            raise ValueError("Monitoring interval should be a number: '%s'" %
                             self.interval)

        cfg = ConfigParser.ConfigParser()
        cfg.add_section('main')
        cfg.set('main', 'interval', self.interval)
        cfg.set('main', 'host', self.host)
        cfg.set('main', 'loglevel', loglevel)
        cfg.set('main', 'username', self.username)

        cfg.add_section('metric')
        cfg.set('metric', 'names', self.metric)

        cfg.add_section('custom')
        for method in self.custom:
            if self.custom[method]:
                cfg.set('custom', method, ','.join(self.custom[method]))

        cfg.add_section('startup')
        for idx, cmd in enumerate(self.startups):
            cfg.set('startup', "cmd%s" % idx, cmd)

        cfg.add_section('shutdown')
        for idx, cmd in enumerate(self.shutdowns):
            cfg.set('shutdown', "cmd%s" % idx, cmd)

        with open(self.path['TEMP_CONFIG'], 'w') as fds:
            cfg.write(fds)

        return self.path['TEMP_CONFIG']

    def install(self, loglevel):
        """Create folder and copy agent and metrics scripts to remote host"""
        logger.info("Installing monitoring agent at %s@%s...", self.username,
                    self.host)

        # create remote temp dir
        cmd = self.python + ' -c "import tempfile; print tempfile.mkdtemp();"'
        logger.info("Creating temp dir on %s", self.host)
        try:
            out, errors, err_code = self.ssh.execute(cmd)
        except:
            logger.error("Failed to install monitoring agent to %s",
                         self.host,
                         exc_info=True)
            return None
        if errors:
            logging.error("[%s] error: '%s'", self.host, errors)
            return None

        if err_code:
            logging.error("Failed to create remote dir via SSH"
                          " at %s@%s, code %s: %s" % (self.username, self.host,
                                                      err_code, out.strip()))
            return None

        remote_dir = out.strip()
        if remote_dir:
            self.path['AGENT_REMOTE_FOLDER'] = remote_dir
        logger.debug("Remote dir at %s:%s", self.host,
                     self.path['AGENT_REMOTE_FOLDER'])

        # Copy agent and config
        agent_config = self.create_agent_config(loglevel)
        try:
            self.ssh.send_file(self.path['AGENT_LOCAL_FOLDER'] + '/agent.py',
                               self.path['AGENT_REMOTE_FOLDER'] + '/agent.py')
            self.ssh.send_file(agent_config,
                               self.path['AGENT_REMOTE_FOLDER'] + '/agent.cfg')
        except:
            logger.error("Failed to install agent on %s",
                         self.host,
                         exc_info=True)
            return None
        return agent_config

    def uninstall(self):
        """
        Remove agent's files from remote host
        """
        if self.session:
            self.session.send("stop\n")
            self.session.close()
        fhandle, log_filename = tempfile.mkstemp('.log',
                                                 "agent_" + self.host + "_")
        os.close(fhandle)
        try:
            self.ssh.get_file(self.path['AGENT_REMOTE_FOLDER'] + "_agent.log",
                              log_filename)
            self.ssh.rm_r(self.path['AGENT_REMOTE_FOLDER'])
        except:
            logger.error("Exception while uninstalling agent", exc_info=True)

        logger.info("Removing agent from: %s@%s...", self.username, self.host)
        return log_filename


class MonitoringCollector(object):
    """Aggregate data from several collectors"""

    def __init__(self):
        self.config = None
        self.default_target = None
        self.agents = []
        self.agent_sessions = []
        self.filter_conf = {}
        self.listeners = []
        self.first_data_received = False
        self.send_data = []
        self.artifact_files = []
        self.inputs, self.outputs, self.excepts = [], [], []
        self.filter_mask = defaultdict(str)
        self.ssh_timeout = 5
        self.load_start_time = None

    def add_listener(self, obj):
        self.listeners.append(obj)

    def prepare(self):
        """Prepare for monitoring - install agents etc"""

        # Parse config
        agent_config = []
        if self.config:
            [agent_config, self.filter_conf] = self.getconfig(
                self.config, self.default_target)
        loglevel = Config(self.config).loglevel()

        logger.debug("filter_conf: %s", self.filter_conf)

        # Filtering
        for host in self.filter_conf:
            self.filter_mask[host] = []
        logger.debug("Filter mask: %s", self.filter_mask)

        # Creating agent for hosts
        logger.debug('Creating agents')
        for adr in agent_config:
            logger.debug('Creating agent: %s', adr)
            agent = AgentClient(adr, timeout=self.ssh_timeout)
            logger.debug('Install monitoring agent. Host: %s', agent.host)
            agent_config = agent.install(loglevel)
            if agent_config:
                self.agents.append(agent)
                self.artifact_files.append(agent_config)

    def start(self):
        """Start N parallel agents"""
        [agent.start() for agent in self.agents]

    def poll(self):
        """Poll agents for data"""
        for agent in self.agents:
            block = agent.read_maybe()
            if not block:
                continue
            lines = block.split("\n")

            for data in lines:
                logger.debug("Got data from agent: %s", data.strip())
                self.send_data.append(
                    self.filter_unused_data(
                        self.filter_conf, self.filter_mask, data
                    )
                )
                logger.debug("Data after filtering: %s", self.send_data)

        if not self.first_data_received and self.send_data:
            self.first_data_received = True
            logger.info("Monitoring received first data")
        else:
            self.send_collected_data()

        return len(self.outputs)

    def stop(self):
        """Shutdown agents"""
        logger.debug("Uninstalling monitoring agents")
        for agent in self.agents:
            self.artifact_files.append(agent.uninstall())

    def send_collected_data(self):
        """sends pending data set to listeners"""
        [listener.monitoring_data(self.send_data) for listener in self.listeners]
        self.send_data = []

    def get_host_config(self, host, target_hint):

        default = {
            'System': 'csw,int',
            'CPU': 'user,system,iowait',
            'Memory': 'free,cached,used',
            'Disk': 'read,write',
            'Net': 'recv,send,rx,tx',
        }

        default_metric = ['CPU', 'Memory', 'Disk', 'Net']

        names = defaultdict()

        hostname = host.get('address').lower()
        if hostname == '[target]':
            if not target_hint:
                raise ValueError("Can't use [target] keyword with "
                                 "no target parameter specified")
            logger.debug("Using target hint: %s", target_hint)
            hostname = target_hint.lower()
        stats = []
        startups = []
        shutdowns = []
        custom = {'tail': [], 'call': [], }
        metrics_count = 0
        for metric in host:
            # known metrics
            if metric.tag in default.keys():
                metrics_count += 1
                metr_val = default[metric.tag].split(',')
                if metric.get('measure'):
                    metr_val = metric.get('measure').split(',')
                for elm in metr_val:
                    if not elm:
                        continue
                    stat = "%s_%s" % (metric.tag, elm)
                    stats.append(stat)
                    agent_name = self.get_agent_name(metric.tag, elm)
                    if agent_name:
                        names[agent_name] = 1
            # custom metric ('call' and 'tail' methods)
            elif (str(metric.tag)).lower() == 'custom':
                metrics_count += 1
                isdiff = metric.get('diff')
                if not isdiff:
                    isdiff = 0
                stat = "%s:%s:%s" % (base64.b64encode(metric.get('label')),
                                     base64.b64encode(metric.text), isdiff)
                stats.append('Custom:' + stat)
                custom[metric.get('measure', 'call')].append(stat)
            elif (str(metric.tag)).lower() == 'startup':
                startups.append(metric.text)
            elif (str(metric.tag)).lower() == 'shutdown':
                shutdowns.append(metric.text)

        logger.debug("Metrics count: %s", metrics_count)
        logger.debug("Host len: %s", len(host))
        logger.debug("keys: %s", host.attrib.keys())
        logger.debug("values: %s", host.attrib.values())

        # use default metrics for host
        if metrics_count == 0:
            for metric in default_metric:
                metr_val = default[metric].split(',')
                for elm in metr_val:
                    stat = "%s_%s" % (metric, elm)
                    stats.append(stat)
                    agent_name = self.get_agent_name(metric, elm)
                    if agent_name:
                        names[agent_name] = 1

        metric = ','.join(names.keys())
        if not metric and not custom:
            metric = "cpu-stat"

        return {
            'metric': metric,
            'interval': host.get('interval', 1),
            'priority': host.get('priority', 0),
            'port': int(host.get('port', 22)),
            'python': host.get('python', '/usr/bin/env python2'),
            'username': host.get('username', getpass.getuser()),
            'custom': custom,
            'host': hostname,
            'startups': startups,
            'shutdowns': shutdowns,

            # XXX: should be separate?
            'stats': {hostname: stats},
        }

    def getconfig(self, filename, target_hint):
        """Prepare config data"""

        try:
            tree = parse_xml(filename)
        except IOError as exc:
            logger.error("Error loading config: %s", exc)
            raise RuntimeError("Can't read monitoring config %s" % filename)

        hosts = tree.findall('Host')
        config = []

        filter_obj = defaultdict(str)
        for host in hosts:
            host_config = self.get_host_config(host, target_hint)
            # XXX: why stats should be separated?
            filter_obj.update(host_config.pop('stats'))
            config.append(host_config)

        return [config, filter_obj]

    def filtering(self, mask, filter_list):
        """Filtering helper"""
        host = filter_list[0]
        initial = [0, 1]
        res = []
        if mask[host]:
            keys = initial + mask[host]
            for key in keys:
                try:
                    res.append(filter_list[key])
                except IndexError:
                    logger.warn("Problems filtering data: %s with %s", mask,
                                len(filter_list))
                    return None
        return ';'.join(res)

    def filter_unused_data(self, filter_conf, filter_mask, data):
        """Filter unselected metrics from data"""
        logger.debug("Filtering data: %s", data)
        out = ''
        # Filtering data
        keys = data.rstrip().split(';')
        if re.match('^start;', data):  # make filter_conf mask
            host = keys[1]
            for i in range(3, len(keys)):
                if keys[i] in filter_conf[host]:
                    filter_mask[host].append(i - 1)
            logger.debug("Filter mask: %s", filter_mask)
            out = 'start;'
            out += self.filtering(filter_mask, keys[1:]).rstrip(';') + '\n'
        elif re.match('^\[debug\]', data):  # log debug output
            logger.debug('agent debug: %s', data.rstrip())
        else:
            # if we are in start_test() phase, check data's timestamp with load_start_time
            # and skip data collected before load_start_time
            if self.load_start_time is not None:
                try:
                    if int(keys[1]) >= self.load_start_time:
                        filtered = self.filtering(filter_mask, keys)
                        if filtered:
                            out = filtered + '\n'  # filtering values
                except IndexError:
                    pass
        return out

    def get_agent_name(self, metric, param):
        """Resolve metric name"""
        depend = {
            'CPU': {
                'idle': 'cpu-stat',
                'user': 'cpu-stat',
                'system': 'cpu-stat',
                'iowait': 'cpu-stat',
                'nice': 'cpu-stat'
            },
            'System': {
                'la1': 'cpu-la',
                'la5': 'cpu-la',
                'la15': 'cpu-la',
                'csw': 'cpu-stat',
                'int': 'cpu-stat',
                'numproc': 'cpu-stat',
                'numthreads': 'cpu-stat',
            },
            'Memory': {
                'free': 'mem',
                'used': 'mem',
                'cached': 'mem',
                'buff': 'mem',
            },
            'Disk': {
                'read': 'disk',
                'write': 'disk',
            },
            'Net': {
                'recv': 'net',
                'send': 'net',
                'tx': 'net-tx-rx',
                'rx': 'net-tx-rx',
                'retransmit': 'net-retrans',
                'estab': 'net-tcp',
                'closewait': 'net-tcp',
                'timewait': 'net-tcp',
            }
        }
        if depend[metric][param]:
            return depend[metric][param]
        else:
            return ''


class StdOutPrintMon(MonitoringDataListener):
    """Simple listener, writing data to stdout"""

    def __init__(self):
        MonitoringDataListener.__init__(self)

    def monitoring_data(self, data_list):
        [sys.stdout.write(data) for data in data_list]


class MonitoringDataDecoder(object):
    """The class that serves converting monitoring data lines to dict"""
    NA = 'n/a'

    def __init__(self):
        self.metrics = {}

    def decode_line(self, line):
        """convert mon line to dict"""
        is_initial = False
        data_dict = {}
        data = line.strip().split(';')
        timestamp = -1
        if data[0] == 'start':
            data.pop(0)  # remove 'start'
            host = data.pop(0)
            if not data:
                logger.warn("Wrong mon data line: %s", line)
            else:
                timestamp = data.pop(0)
                self.metrics[host] = []
                for metric in data:
                    if metric.startswith("Custom:"):
                        metric = base64.standard_b64decode(metric.split(':')[
                            1])
                    self.metrics[host].append(metric)
                    data_dict[metric] = self.NA
                    is_initial = True
        else:
            host = data.pop(0)
            timestamp = data.pop(0)

            if host not in self.metrics.keys():
                raise ValueError("Host %s not in started metrics: %s" %
                                 (host, self.metrics))

            if len(self.metrics[host]) != len(data):
                raise ValueError("Metrics len and data len differs: %s vs %s" %
                                 (len(self.metrics[host]), len(data)))

            for metric in self.metrics[host]:
                data_dict[metric] = data.pop(0)

        logger.debug("Decoded data %s: %s", host, data_dict)
        return host, data_dict, is_initial, timestamp

# FIXME: 3 synchronize times between agent and collector better
